// grab chart holder DOM element
const defaultChart = document.getElementById('defaultChart');
const inboundChart = document.getElementById('inboundChart');
const outboundChart = document.getElementById('outboundChart');
const canvas = document.getElementById("mapCanvas");
const ctx = canvas && canvas.getContext("2d");
const { LineChart, defaultColors } = Charts;
const tickBuckets = 20;
const dataUpdateFrequecySeconds = 3;
const getReqHeader = () => {
  return new Headers({
    'Content-Type': 'application/json',
  })
};

const getDefaultDataset = () => {
  return {
    label: 'Total',
    backgroundColors: [defaultColors[0]],
    data: [0],
    id: 'Total',
  }
}

const getDefaultData = () => {
  return {
    labels: [0],
    datasets: [
      getDefaultDataset(),
    ],
  }
};

const getChartOptions = (metadata=null) => {
  return {
    scales: {
        x: {
            title: metadata.x_title || '',
        },
        y: {
            // title: 'Figures',
            formatter: axisValue => {
            return getValueWithProperUnitFromBytes(axisValue);
            },
            stacked: false,
        },
    },
    legendClickable: true,
    containerResizable: true,
    curve: 'curveBasis',
  };
}

let inboundData = getDefaultData();
let outboundData = getDefaultData();
const sessionDetailData = getDefaultData();

const tmpTotalDataset = getDefaultDataset();

const defaultChartOptions = getChartOptions({ x_title: '' });
const inboundChartOptions = getChartOptions({ x_title: 'Inbound'});
const outboundChartOptions = getChartOptions({ x_title: 'Outbound'});
let intervalFunc;

const plotChart = (chartDOM, data, options) => {
  chartDOM.innerHTML='';
  // initialize the chart
  const chart = new Charts.LineChart(chartDOM, {
    data: data,
    options: options,
  });
  return chart;
}

const activeTransfersStatuses = ['running', 'paused', 'queued', 'submitted', ];
const getAllActiveTransfers = datetime => {
  const successCallback = result => {
    const activeTransferArr = Array.isArray(result) ? result.filter(transfer => activeTransfersStatuses.includes(transfer.status)) : [];
    return activeTransferArr;
  };

  return fetch(`/aspera/console/api/transfers?${formatQueryParams(datetime)}&filter1=status&comp1=neq&val1=completed&filter2=status&comp2=neq&val2=error`, {
    credentials: "same-origin",
    headers: getReqHeader(),
  })
  .then(response => response.json())
  .then(successCallback)
  .catch(err => {
    console.log(err);
    return [];
  })
}

const getTransferDetail = job_id => {
  const successCallback = result => result;
  return fetch(`/aspera/console/api/transfers/${job_id}`, {
    credentials: "same-origin",
    headers: getReqHeader(),
  })
  .then(response => response.json())
  .then(successCallback)
  .catch(err => {
    console.log(err);
    return {};
  })
}

const getInboundTransfers = (nodes_ip_arr, transfers) => transfers.filter(transfer => nodes_ip_arr.includes(transfer.destination));
const getInboundTransfersWithNodeName = (node_name, transfers) => transfers.filter(transfer => transfer.destination_description === node_name);
const getOutboundTransfers = (nodes_ip_arr, transfers) => transfers.filter(transfer => nodes_ip_arr.includes(transfer.source));
const getOutboundTransfersWithNodeName = (node_name, transfers) => transfers.filter(transfer => transfer.source_description === node_name);
const getSingleNodeTransfers = (node_ip, transfers) => transfers.filter(transfer => transfer.source === node_ip || transfer.destination === node_ip);
const getTransfersIds = transfers => transfers.map(transfer => transfer.id);
const getTotalRateFromTrasfersDetailList = transfersDetailList => transfersDetailList.reduce((accumulator, transferDetail) => accumulator + transferDetail.last_calculated_rate_bps, 0);
const moreThanOneTransferHasRatelargerThanZero = transfersDetailList => transfersDetailList.filter(transferDetail => transferDetail.last_calculated_rate_bps > 0).length > 1;
const addZeroOrNot = timeValue => timeValue < 10 ? `0${timeValue}` : timeValue;
const cloneDeep = value => JSON.parse(JSON.stringify(value));
const getTransferSelectItemText = info_hash => getTransferUniqueLabel(info_hash)
const getTransferUniqueLabel = info_hash => `${info_hash.name} (${info_hash.source_description} -> ${info_hash.destination_description}) (${info_hash.id})`
const getMapNodeDomId = text => text.replace(/[^0-9A-Za-z]/g , '_');
const getTime24HrsAgoInMillSecs = () => new Date(Date.now() - (24 * 60 * 60 * 1000) + (10 * 60 * 1000));
const getMbFromBytes = value => (value/1000/1000).toFixed(2);
const getKbFromBytes = value => (value/1000).toFixed(1);
const getValueWithProperUnitFromBytes = (value=0) => formatBytes(value, 1, true)
const formatBytes = (bytes, decimals=1, add_unit=false) => {
  if (bytes == 0) return '0 bps';
  const k = 1000,
        dm = decimals <= 0 ? 0 : decimals,
        sizes = ['bps', 'Kbps', 'Mbps', 'Gbps', 'Tbps', 'Pbps', 'Ebps', 'Zbps', 'Ybps'];
  const i = Math.floor(Math.log(bytes) / Math.log(k));
  const value = parseFloat((bytes / Math.pow(k, i))).toFixed(dm);
  return add_unit ? `${value} ${sizes[i]}` : value;
}
const resetInboundData = () => inboundData = getDefaultData();
const resetOutboundData = () => outboundData = getDefaultData();
let timerTimestamp = new Date(Date.now());
const getCurrTimestamp = () => new Date(Date.now());
const chartShouldBeRedrawn = () => (getCurrTimestamp() - timerTimestamp) > ( 1 * 60 * 1000);
const tabIsNotDisplayed = () => !! (document.hidden || document.mozHidden || document.msHidden || document.webkitHidden);

const formatQueryParams = datetime => {
  const year = datetime.getUTCFullYear();
  const month = addZeroOrNot(datetime.getUTCMonth() + 1);
  const date = addZeroOrNot(datetime.getUTCDate());
  const hour = addZeroOrNot(datetime.getUTCHours());
  const minute = addZeroOrNot(datetime.getUTCMinutes());
  const second = addZeroOrNot(datetime.getUTCSeconds());
  return `from=${year}-${month}-${date}%20${hour}:${minute}:${second}`;
}

const hideTotalLineIfOnlyOneTransfer = (dataMap, allTransfersDetail) => {
  dataMap.datasets = dataMap.datasets.filter(dataset => dataset.id !== 'Total');
  (dataMap.datasets.length > 1 || allTransfersDetail.length === 0) && dataMap.datasets.push(cloneDeep(tmpTotalDataset));
}

let currBandwidthNodesArr = [];
const changeBandwidthLegendNodes = (type=null, node_data_or_node_ips=null) => {
  clearInterval(intervalFunc);
  jQuery(function($) {
  const nodeInfo = [];
  if (type === 'node_ips') { // when user click a transfer line on map
    currBandwidthNodesArr.forEach(nodeMetadata => $(`#map_node_${getMapNodeDomId(nodeMetadata.name)} button span`).removeClass('selected_node_on_edit_map'));
    currBandwidthNodesArr = [];
    nodeInfo.push({
      ip: node_data_or_node_ips.x1_ip,
      name: node_data_or_node_ips.x1_name,
      node_id: node_data_or_node_ips.x1_node_id,
      managed: node_data_or_node_ips.x1_node_managed,
    })
    nodeInfo.push({
      ip: node_data_or_node_ips.x2_ip,
      name: node_data_or_node_ips.x2_name,
      node_id: node_data_or_node_ips.x2_node_id,
      managed: node_data_or_node_ips.x2_node_managed,
    })
    nodeInfo.forEach(nodeMetadata => {
      if (currBandwidthNodesArr.find(metadata => metadata.ip === nodeMetadata.ip)){
        currBandwidthNodesArr = currBandwidthNodesArr.filter(metadata => metadata.ip !== nodeMetadata.ip);
      } else {
        currBandwidthNodesArr.push(nodeMetadata);
      }
    })
  } else { // when user click a node on map
    const wasSelectingTransferLine = currBandwidthNodesArr.every(nodeMetadata => !$(`#map_node_${getMapNodeDomId(nodeMetadata.name)} button span`).hasClass('selected_node_on_edit_map'));
    if (wasSelectingTransferLine){
      currBandwidthNodesArr = [];
    }
    nodeInfo.push({
      ip: nodes.find(node => node.node_name === node_data_or_node_ips.name).node_address,
      name: node_data_or_node_ips.name,
      node_id: node_data_or_node_ips.id,
      managed: node_data_or_node_ips.type === 'managed',
    })
    nodeInfo.forEach(nodeMetadata => {
      if (currBandwidthNodesArr.find(metadata => metadata.ip === nodeMetadata.ip)){
        currBandwidthNodesArr = currBandwidthNodesArr.filter(metadata => metadata.ip !== nodeMetadata.ip);
        $(`#map_node_${getMapNodeDomId(nodeMetadata.name)} button span`).removeClass('selected_node_on_edit_map');
      } else {
        currBandwidthNodesArr.push(nodeMetadata);
        $(`#map_node_${getMapNodeDomId(nodeMetadata.name)} button span`).addClass('selected_node_on_edit_map');
      }
    })
  }

  inboundData = getDefaultData();
  outboundData = getDefaultData();

  setTimeout(async () => {
    if (currBandwidthNodesArr.length === 0){
      const charts = await plotDashboardChart();
      intervalFunc = setInterval(function(){ updateDashboardChart(null, charts)}, dataUpdateFrequecySeconds * 1000);
    } else {
      const charts = await plotDashboardChart(currBandwidthNodesArr.map(node => node.ip));
      intervalFunc = setInterval(function(){ updateDashboardChart(currBandwidthNodesArr.map(node => node.ip), charts)}, dataUpdateFrequecySeconds * 1000);
    }
  }, 300);

  // update bandwidth title
  if (currBandwidthNodesArr.length === 0) {
    $('#bandwidth-title').html(`<a href="/aspera/console/bandwidth/show_all"> Bandwidth </a>`);
  } else if (currBandwidthNodesArr.length >= 3) {
    $('#bandwidth-title').html(`<a href="/aspera/console/bandwidth/show_all"> Bandwidth </a> for multiple nodes`);
  } else {
    let nodesTitle = '';
    currBandwidthNodesArr.forEach(nodeMetadata => nodesTitle += nodeMetadata.managed ? `<a href="/aspera/console/nodes/${nodeMetadata.node_id}">${nodeMetadata.name}</a> ` : `${nodeMetadata.name} `);
    $('#bandwidth-title').html(`<a href="/aspera/console/bandwidth/show_all"> Bandwidth </a> for ${nodesTitle}`);
  }
})
}

const getNodeTooltipText = (node_name, inboundTransfers, outboundTransfers) => {
  let tooltipText =
    `${node_name}` +
    '<br>' +
    '---------------' +
    '<br>' +
    `Incoming: ${getValueWithProperUnitFromBytes(getTotalRateFromTrasfersDetailList(inboundTransfers))}` +
    '<br>';
    inboundTransfers.forEach(transfer => tooltipText += `${transfer.name} (${transfer.source_description} ->): ${getValueWithProperUnitFromBytes(transfer.last_calculated_rate_bps)}` + '<br>');
  tooltipText +=
  '---------------' +
  '<br>';
  tooltipText +=
  `Outgoing: ${getValueWithProperUnitFromBytes(getTotalRateFromTrasfersDetailList(outboundTransfers))}` +
  '<br>';
  outboundTransfers.forEach(transfer => tooltipText += `${transfer.name} (-> ${transfer.destination_description}): ${getValueWithProperUnitFromBytes(transfer.last_calculated_rate_bps)}` + '<br>');
  return tooltipText;
}

let mapSelectedNode;
const drawMapNodes = allActiveTransfers => {
  const isTheNodeForEditMapPage = map_node => map_node.selected === true;
  _map_nodes.forEach(async map_node => {
    const inboundTransfersOfSingleMapNode = getInboundTransfersWithNodeName(map_node.name, allActiveTransfers);
    const outboundTransfersOfSingleMapNode = getOutboundTransfersWithNodeName(map_node.name, allActiveTransfers);
    const inboundTransferListOfSingleMapNode = await Promise.all(getTransfersIds(inboundTransfersOfSingleMapNode).map(transferId => getTransferDetail(transferId)));
    const outboundTransferListOfSingleMapNode = await Promise.all(getTransfersIds(outboundTransfersOfSingleMapNode).map(transferId => getTransferDetail(transferId)));
    const marginLeftPx = map_node.x * canvas.width;
    const marginTopPx = map_node.y * canvas.height;
    const mapNodeDom = document.getElementById(`map_node_${getMapNodeDomId(map_node.name)}`);
    const tooltipText = getNodeTooltipText(map_node.name, inboundTransferListOfSingleMapNode, outboundTransferListOfSingleMapNode);
    mapSelectedNode = mapSelectedNode || (isTheNodeForEditMapPage(map_node) ? map_node : null);
    if (!mapNodeDom){
      const span = document.createElement("SPAN");
      span.setAttribute("class", `bx--tag bx--tag--${ map_node.type === 'managed' ? map_node.status === 'dead' ? 'magenta' : 'green' : 'teal' } ${isEditMapPage() && isTheNodeForEditMapPage(map_node) && 'selected_node_on_edit_map'}`);
      span.style.width = "1rem";
      span.style.height = "1rem";
      span.style.padding = "0";
      span.style.border = 'solid #000';
      span.style.borderWidth = '1px';
      const button = document.createElement("BUTTON");
      button.setAttribute("class", "bx--tooltip__trigger bx--tooltip--a11y bx--tooltip__trigger--definition bx--tooltip--bottom bx--tooltip--align-end")
      if (!isEditMapPage()){
        if (isDashboardPage()){
          map_node.type === "managed" && map_node.status !== 'dead' ? button.onclick = () => changeBandwidthLegendNodes('node_name', map_node) : button.style.cursor = 'default';
        } else {
          map_node.type === "managed" ? button.onclick = () => window.location.replace(`/aspera/console/nodes/${map_node.id}`) : button.style.cursor = 'default';
        }
      } else {
        button.style.cursor = isTheNodeForEditMapPage(map_node) ? 'all-scroll' : 'unset';
      }
      button.appendChild(span);
      const tooltipTextDom = document.createElement("DIV");
      tooltipTextDom.setAttribute("class", "bx--assistive-text");
      tooltipTextDom.setAttribute("role", "tooltip");
      tooltipTextDom.innerHTML = tooltipText;
      const tooltipContainer = document.createElement("DIV");
      tooltipContainer.setAttribute("id", `map_node_${getMapNodeDomId(map_node.name)}`);
      tooltipContainer.setAttribute("class", "bx--tooltip--definition bx--tooltip--a11y");
      tooltipContainer.appendChild(button);
      tooltipContainer.appendChild(tooltipTextDom);
      tooltipContainer.style.position = "absolute";
      tooltipContainer.style.left = `${marginLeftPx}px`;
      tooltipContainer.style.top = `${marginTopPx}px`;
      isEditMapPage() && isTheNodeForEditMapPage(map_node) && dragElement(tooltipContainer);
      document.getElementById("new_dashboard_map").appendChild(tooltipContainer);
    } else {
      mapNodeDom.childNodes[1].innerHTML = tooltipText;
    }
  })
}

const drawMapLines = allTransfersDetails => {
  jQuery(function($) {
    const dataForDrawlingLine = {};
    const updateDataForDrawingLine = () => {
      const aggregateNodeIdAndIsManagedToTD = (transferDetail, source_node, destination_node) => {
        transferDetail.source_node_id = source_node.id;
        transferDetail.source_node_managed = source_node.type === 'managed';
        transferDetail.destination_node_id = destination_node.id;
        transferDetail.destination_node_managed = destination_node.type === 'managed';
        return transferDetail;
      }
      const getTransferDetailKey = (description_1, description_2) => `${description_1}_${description_2}`;

      allTransfersDetails.forEach(async transferDetail => {
        const source_node_in_map_nodes = _map_nodes.find(map_node => map_node.name === transferDetail.source_description);
        const destination_node_in_map_nodes = _map_nodes.find(map_node => map_node.name === transferDetail.destination_description);
        if (source_node_in_map_nodes && destination_node_in_map_nodes){
          const keyAsSourceFirst = getTransferDetailKey(transferDetail.source_description, transferDetail.destination_description);
          const keyAsDestinationFirst = getTransferDetailKey(transferDetail.destination_description, transferDetail.source_description);

          if (keyAsSourceFirst in dataForDrawlingLine) {
            transferDetail = aggregateNodeIdAndIsManagedToTD(transferDetail, source_node_in_map_nodes, destination_node_in_map_nodes);
            dataForDrawlingLine[keyAsSourceFirst].push(transferDetail);
          } else if (keyAsDestinationFirst in dataForDrawlingLine) {
            transferDetail = aggregateNodeIdAndIsManagedToTD(transferDetail, source_node_in_map_nodes, destination_node_in_map_nodes);
            dataForDrawlingLine[keyAsDestinationFirst].push(transferDetail);
          } else {
            transferDetail = aggregateNodeIdAndIsManagedToTD(transferDetail, source_node_in_map_nodes, destination_node_in_map_nodes);
            dataForDrawlingLine[keyAsSourceFirst] = [transferDetail];
          }
        }
      })
    }
    const drawLinesWithData = () => {
      const lineHoverDataArr = [];
      for (let transferSourceAndDestination in dataForDrawlingLine){
        const transferDetailList = dataForDrawlingLine[transferSourceAndDestination];
        const pos1 = $(`#map_node_${getMapNodeDomId(transferDetailList[0].source_description)}`).position();
        const pos2 = $(`#map_node_${getMapNodeDomId(transferDetailList[0].destination_description)}`).position();
        const pos1_left = pos1 && pos1.left + 10;
        const pos1_top = pos1 && pos1.top + 10;
        const pos2_left = pos2 && pos2.left + 10;
        const pos2_top = pos2 && pos2.top + 10;
        ctx.moveTo(pos1_left, pos1_top);
        ctx.lineTo(pos2_left, pos2_top);
        let transferMsg = '';
        transferDetailList.forEach(transferDetail => {
          transferMsg += `${transferDetail.name} (${transferDetail.source_description} -> ${transferDetail.destination_description}):  ${getValueWithProperUnitFromBytes(transferDetail.last_calculated_rate_bps)}`;
          transferMsg += '<br>';
        })
        const metadataTwoNode = {
          x1: pos1_left,
          y1: pos1_top,
          x2: pos2_left,
          y2: pos2_top,
          x1_ip: transferDetailList[0].source,
          x1_name: transferDetailList[0].source_description,
          x1_node_id: transferDetailList[0].source_node_id,
          x1_node_managed: transferDetailList[0].source_node_managed,
          x2_ip: transferDetailList[0].destination,
          x2_name: transferDetailList[0].destination_description,
          x2_node_id: transferDetailList[0].destination_node_id,
          x2_node_managed: transferDetailList[0].destination_node_managed,
        }
        lineHoverDataArr.push({
          metadataTwoNode,
          transferMsg,
          line_id: transferSourceAndDestination,
        })
      }
      lineHover(lineHoverDataArr);
      ctx.stroke();
    }
    ctx.clearRect(0, 0, canvas.width, canvas.height);
    ctx.beginPath();
    ctx.strokeStyle = "#0ba1d8";
    ctx.lineWidth = 1.5;
    updateDataForDrawingLine();
    drawLinesWithData();
  })
}

const adjustDataMapSize = (queryTimestamp, dataMap) => {
  if (dataMap.labels.length >= tickBuckets) {
    dataMap.labels.shift()
    dataMap.datasets.forEach(dataset => dataset.data.shift())
  }
  const now = getTime24HrsAgoInMillSecs();
  dataMap.labels.push(`${now.getHours()}:${now.getMinutes()}:${now.getSeconds()}`)
}

const colorLuminance = (hex, lum) => {

	// validate hex string
	hex = String(hex).replace(/[^0-9a-f]/gi, '');
	if (hex.length < 6) {
		hex = hex[0]+hex[0]+hex[1]+hex[1]+hex[2]+hex[2];
	}
	lum = lum || 0;

	// convert to decimal and change luminosity
	let rgb = "#", c;
	for (let i = 0; i < 3; i++) {
		c = parseInt(hex.substr(i*2,2), 16);
		c = Math.round(Math.min(Math.max(0, c + (c * lum)), 255)).toString(16);
		rgb += ("00"+c).substr(c.length);
	}
	return rgb;
}

const updateDataMapForSingleTransferDetailOrSingleNode = (dataMap, metadata) => {
  const isNodeDetail = !!metadata.node_address;
  const pk = isNodeDetail ? metadata.node_address : metadata.id;
  const matchedDataset = dataMap.datasets.find(dataset => dataset.id === pk);
  let tmpDataForTotal;
  const datasetsLength = dataMap.datasets.find(dataset => dataset.id === 'Total') ? dataMap.datasets.length : dataMap.datasets.length + 1;
  let dataLength = dataMap.datasets[0].data.length;
  if(!matchedDataset){
    const prefixEmptyData = Array(dataLength).fill(0);
    prefixEmptyData.push(metadata.last_calculated_rate_bps || 0);
    const newDataset = {
      label: isNodeDetail ? metadata.node_address_description : getTransferUniqueLabel(metadata),
      backgroundColors: [colorLuminance(defaultColors[datasetsLength % 8], Math.floor(datasetsLength/8) * -0.15)],
      data: prefixEmptyData,
      id: pk,
    };
    tmpDataForTotal = newDataset.data;
    dataMap.datasets.push(newDataset);
  } else {
    matchedDataset.data.push(metadata.last_calculated_rate_bps || 0);
    tmpDataForTotal = matchedDataset.data;
  }

  if (pk === 'Total') {
    tmpTotalDataset.data = cloneDeep(tmpDataForTotal);
  }
}

let prevInboundTotalRate = 0;
let prevOutboundTotalRate = 0;

const updateDataForSingleLine = async (now, dataMap, transfersIds) => {
  adjustDataMapSize(now, dataMap);
  const getTransferDetailArrow = async transferId => await getTransferDetail(transferId);
  const allTransfersDetail = await Promise.all(transfersIds.map(getTransferDetailArrow));
  const totalRate = await getTotalRateFromTrasfersDetailList(allTransfersDetail);
  updateDataMapForSingleTransferDetailOrSingleNode(dataMap, { id: 'Total', last_calculated_rate_bps: totalRate || 0 });
  return dataMap;
}

const updateDataForMultipleLines = async (now, dataMap, transfersIds, inOrOut=null, groupByNode=null) => {
  const getTransferDetailArrow = async transferId => await getTransferDetail(transferId);
  const allTransfersDetail = await Promise.all(transfersIds.map(getTransferDetailArrow));
  let totalRate = transfersIds.length === 0 ? 0 : await getTotalRateFromTrasfersDetailList(allTransfersDetail);
  
  if (inOrOut === 'inbound') {
    prevInboundTotalRate = cloneDeep(totalRate) || prevInboundTotalRate;
  } else {
    prevOutboundTotalRate = cloneDeep(totalRate) || prevOutboundTotalRate;
  }
  adjustDataMapSize(now, dataMap);

  const type = inOrOut === 'inbound' ? 'destination' : 'source';
  const typeDescription = inOrOut === 'inbound' ? 'destination_description' : 'source_description';
  const allTransfersDetailGroupByNode = [];
  allTransfersDetail.forEach(transferDetail => {
    const TDinGBNode = allTransfersDetailGroupByNode.find(transferDetailGroupByNode => transferDetailGroupByNode.node_address === transferDetail[type])
    if (!TDinGBNode){
      const dataObj = {};
      dataObj['node_address'] = transferDetail[type];
      dataObj['node_address_description'] = transferDetail[typeDescription];
      for (let key in transferDetail){
        dataObj[key] = transferDetail[key];
      }
      allTransfersDetailGroupByNode.push(dataObj);
    } else {
      TDinGBNode.last_calculated_rate_bps += transferDetail.last_calculated_rate_bps;
    }
  });

  let node_ips;
  if (groupByNode) {
    allTransfersDetailGroupByNode.forEach(nodeDetail => updateDataMapForSingleTransferDetailOrSingleNode(dataMap, nodeDetail));
    node_ips = allTransfersDetailGroupByNode.map(transferOfSingleNode => transferOfSingleNode.node_address);
  } else {
    allTransfersDetail.forEach(transferDetail => updateDataMapForSingleTransferDetailOrSingleNode(dataMap, transferDetail));
    node_ips = allTransfersDetail.map(transferDetail => transferDetail.id);
  }

  // append 0 as rate for transfer that completed until 0s reached chart bucket size
  let tmpDatasets = cloneDeep(dataMap.datasets);
  dataMap.datasets.forEach(dataset => {
    if (!node_ips.includes(dataset.id) && dataset.id !== 'Total'){
      if(dataset.data.every(datapoint => datapoint === 0)){
        tmpDatasets = tmpDatasets.filter(tmpDataset => tmpDataset.id !== dataset.id)
      } else{
        tmpDatasets.find(tmpDataset => tmpDataset.id === dataset.id).data.push(0)
      }
    }
  })
  dataMap.datasets = tmpDatasets;
  //--------------------
  if (isNaN(totalRate)){
    totalRate = inOrOut === 'inbound' ? prevInboundTotalRate : prevOutboundTotalRate;
  }
  updateDataMapForSingleTransferDetailOrSingleNode(dataMap, { id: 'Total', last_calculated_rate_bps: totalRate });
  return dataMap;
}

const getDashboardChartData = async (now, allActiveTransfers, param_node_address_arr=null) => {
  // call all transfer api and transfer detail api, add the results, prep the dataset.
  const managedNodes = param_node_address_arr ? param_node_address_arr : (nodes && nodes.map(node => node.node_address));
  const inboundTransfersIds = getTransfersIds(getInboundTransfers(managedNodes, allActiveTransfers));
  const outboundTransfersIds = getTransfersIds(getOutboundTransfers(managedNodes, allActiveTransfers));
  const inboundDataResolve = await updateDataForMultipleLines(now, inboundData, inboundTransfersIds, 'inbound', true);
  const outboundDataResolve = await updateDataForMultipleLines(now, outboundData, outboundTransfersIds, 'outbound', true);
  return {
    inboundDataResolve: inboundDataResolve,
    outboundDataResolve: outboundDataResolve,
  }
}

const plotDashboardChart = async (param_node_address_arr=null) => {
  const now = getTime24HrsAgoInMillSecs();
  const allActiveTransfers = await getAllActiveTransfers(now) || [];
  const dataResolve = await getDashboardChartData(now, allActiveTransfers, param_node_address_arr);
  const inChart = plotChart(inboundChart, dataResolve.inboundDataResolve, inboundChartOptions);
  const outChart = plotChart(outboundChart,  dataResolve.outboundDataResolve, outboundChartOptions);
  updateActivityMap(now, allActiveTransfers);
  return {
    inChart: inChart,
    outChart: outChart,
  }
}

let prevInboundDatasetsIds;
let prevOutboundDatasetsIds;
const chartLegendShouldChange = dataResolve => prevInboundDatasetsIds && prevOutboundDatasetsIds && (JSON.stringify(prevInboundDatasetsIds) !== JSON.stringify(dataResolve.inboundDataResolve.datasets.map(dataset => dataset.id)) || JSON.stringify(prevOutboundDatasetsIds) !== JSON.stringify(dataResolve.outboundDataResolve.datasets.map(dataset => dataset.id)));
const trackChartLegend = dataResolve => {
  prevInboundDatasetsIds = dataResolve.inboundDataResolve.datasets.map(dataset => dataset.id);
  prevOutboundDatasetsIds = dataResolve.outboundDataResolve.datasets.map(dataset => dataset.id);
}

const updateDashboardChart = async (param_node_address_arr=null, charts=null) => {
  if (tabIsNotDisplayed()) {
    return;
  }
  const now = getTime24HrsAgoInMillSecs();
  const allActiveTransfers = await getAllActiveTransfers(now) || [];
  const dataResolve = await getDashboardChartData(now, allActiveTransfers, param_node_address_arr);

  charts.inChart.data = dataResolve.inboundDataResolve;
  charts.outChart.data = dataResolve.outboundDataResolve;

  if (chartLegendShouldChange(dataResolve)){
    clearInterval(intervalFunc);
    const new_charts = await plotDashboardChart(param_node_address_arr);
    intervalFunc = setInterval(function(){ updateDashboardChart(param_node_address_arr, new_charts)}, dataUpdateFrequecySeconds * 1000);
  } else {
    charts.inChart.update();
    charts.outChart.update();
  }
  updateActivityMap(now, allActiveTransfers);
  trackChartLegend(dataResolve);
}

const getNodeDetailChartData = async (type=null) => {
  const now = getTime24HrsAgoInMillSecs();
  const allActiveTransfers = await getAllActiveTransfers(now) || [];
  const node_address_DOM = document.getElementById('node_address_hidden');
  const node_address = node_address_DOM && node_address_DOM.value;
  const managedNodes = type === pathMap.all_bandwidth ? (nodes && nodes.map(node => node.node_address)) : [node_address];
  const transfers = type === pathMap.all_bandwidth ? allActiveTransfers : getSingleNodeTransfers(node_address, allActiveTransfers);
  const inboundTransfersIds = getTransfersIds(getInboundTransfers(managedNodes, transfers));
  const outboundTransfersIds = getTransfersIds(getOutboundTransfers(managedNodes, transfers));
  const inboundDataResolve = await updateDataForMultipleLines(now, inboundData, inboundTransfersIds, 'inbound', false);
  const outboundDataResolve = await updateDataForMultipleLines(now, outboundData, outboundTransfersIds, 'outbound', false);
  return {
    inboundDataResolve: inboundDataResolve,
    outboundDataResolve: outboundDataResolve,
    transfers: transfers,
  }
}

const plotNodeDetailChart = async (type=null) => {
  const dataResolve = await getNodeDetailChartData(type);
  const inChart = plotChart(inboundChart, dataResolve.inboundDataResolve, inboundChartOptions);
  const outChart = plotChart(outboundChart, dataResolve.outboundDataResolve, outboundChartOptions);
  updateTransferSelectOptions(dataResolve.transfers);
  return {
    inChart: inChart,
    outChart: outChart,
  }
}

const updateNodeDetailChart = async (type=null, charts=null) => {
  if (tabIsNotDisplayed()) {
    return;
  }
  const dataResolve = await getNodeDetailChartData(type);
  charts.inChart.data = dataResolve.inboundDataResolve;
  charts.outChart.data = dataResolve.outboundDataResolve;
  if (chartLegendShouldChange(dataResolve)){
    clearInterval(intervalFunc);
    const new_charts = await plotNodeDetailChart(type);
    intervalFunc = setInterval(function(){ updateNodeDetailChart(type, new_charts)}, dataUpdateFrequecySeconds * 1000);
  } else {
    charts.inChart.update();
    charts.outChart.update();
  }
  updateTransferSelectOptions(dataResolve.transfers);
  trackChartLegend(dataResolve);
}

const getSessionDetailChartData = async () => {
  // call all transfer api fitler with node address and transfer detail api, add the results, prep the dataset.
  const now = getTime24HrsAgoInMillSecs();
  const job_id = document.getElementById('job_id_hidden').value;
  const dataResolve = await updateDataForSingleLine(now, sessionDetailData, [job_id])
  return {
    singleDataResolve: dataResolve
  };
}

const plotSessionDetailChart = async () => {
  const dataResolve = await getSessionDetailChartData()
  const singleChart = plotChart(defaultChart, dataResolve.singleDataResolve, defaultChartOptions);
  return {
    singleChart: singleChart,
  }
}

const updateSessionDetailChart = async (charts=null) => {
  if (tabIsNotDisplayed()) {
    return;
  }
  const dataResolve = await getSessionDetailChartData();
  charts.singleChart.data = dataResolve.singleDataResolve;
  charts.singleChart.update();
}

const updateActivityMap = async (datetime=null, transfers=null) => {
  if (tabIsNotDisplayed()) {
    return;
  }
  const now = datetime || getTime24HrsAgoInMillSecs();
  const allActiveTransfers = transfers || await getAllActiveTransfers(now) || [];
  const allTransfersDetails = await Promise.all(allActiveTransfers.map(transfer => getTransferDetail(transfer.id)));
  drawMapNodes(allActiveTransfers);
  !isEditMapPage() && drawMapLines(allTransfersDetails);
}

const updateActivityMapDisplay = (width, height) => {
  setCanvasStyleWithMapDom();
  _map_nodes.forEach(async map_node => {
    const marginLeftPx = map_node.x * (width || canvas.width);
    const marginTopPx = map_node.y * (height || canvas.height);
    const mapNodeDom = document.getElementById(`map_node_${getMapNodeDomId(map_node.name)}`);
    mapNodeDom.style.left = `${marginLeftPx}px`;
    mapNodeDom.style.top = `${marginTopPx}px`;
  });
}

const updateTransferSelectOptions = transfers => {
  jQuery(function($) {
    const currTransfersIdsInSelect = $('#default-select-transfer').siblings().toArray().map(option => parseInt(option.value, 10));
    const allTransfersIds = getTransfersIds(transfers);
    // Remove transfers' status that's not paused or active anymore from select
    currTransfersIdsInSelect.forEach(transferId => !allTransfersIds.includes(transferId) && $(`option[value="${transferId}"]`).remove());
    // Add new transfer to select if not already exists
    transfers.forEach(transfer => {
      if (!currTransfersIdsInSelect.includes(transfer.id)) {
        const option = document.createElement("OPTION");
        option.setAttribute("class", "bx--select-option");
        option.setAttribute("value", `${transfer.id}`);
        option.innerHTML = getTransferSelectItemText(transfer);
        document.getElementById("transfer-select") && document.getElementById("transfer-select").appendChild(option);
      }
    });
    // Disabled select when there is only the default selected hidden option
    if ($('#transfer-select option').length === 1){
      $('#transfer-select').attr('disabled', '');
      $('#policy-select').attr('disabled', '');
    } else {
      $('#transfer-select').removeAttr('disabled');
    }
  });
};

const toggleMapDisplay = () => {
  jQuery(function($) {
    $('#new_dashboard_map').toggle();
  })
}

const hideMapDisplayIfNotVisible = _fasp_node => {
  jQuery(function($) {
    !_fasp_node.map_visible && $('#new_dashboard_map').hide();
  })
}

const pathMap = {
  dashboard: 'dashboard',
  node_edit: 'node_edit',
  node_detail: 'node_detail',
  session_detail: 'session_detail',
  all_bandwidth: 'all_bandwidth',
  edit_map: 'edit_map',
  activity_map: 'activity_map',
}

const getCurrPage = () => {
  const currPathElements = window.location.pathname.split('/');
  const path_dashboard = ['dashboard'];
  const path_session_detail = ['job_steps', 'show'];
  const path_node_edit = ['nodes', 'edit'];
  const path_edit_map = ['nodes', 'edit_map'];
  const path_node_detail = ['nodes'];
  const path_unmanaged_node_edit_map = ['partner_nodes', 'edit_map'];
  const path_unmanaged_node_detail = ['partner_nodes'];
  const path_all_bandwidth = ['bandwidth', 'show_all'];
  const path_activity_map = ['map'];

  const matchPathElements = path_arr => path_arr.every(path => currPathElements.includes(path));

  // checking path order matters here
  if (matchPathElements(path_dashboard)){
    return pathMap.dashboard;
  } else if (matchPathElements(path_node_edit)){
    return pathMap.node_edit;
  } else if (matchPathElements(path_edit_map) || matchPathElements(path_unmanaged_node_edit_map)){
    return pathMap.edit_map;
  } else if (matchPathElements(path_node_detail) || matchPathElements(path_unmanaged_node_detail)){
    return pathMap.node_detail;
  } else if (matchPathElements(path_session_detail)){
    return pathMap.session_detail;
  } else if (matchPathElements(path_all_bandwidth)){
    return pathMap.all_bandwidth;
  } else if (matchPathElements(path_activity_map)){
    return pathMap.activity_map;
  } else {
    return pathMap.dashboard;
  }
}

const isDashboardPage = () => getCurrPage() === pathMap.dashboard;
const isEditMapPage = () => getCurrPage() === pathMap.edit_map;
const isAllBandwidthPage = () => getCurrPage() === pathMap.all_bandwidth;

const setCanvasStyleWithMapDom = () => {
  jQuery(function($) {
    const mapMetadata = calculateCanvasMapWidthAndHeight($('#new_dashboard_map').width());
    $('#new_dashboard_map').height(mapMetadata.height);
    canvas.setAttribute('width', mapMetadata.width);
    canvas.setAttribute('height', mapMetadata.height);
  })
}

const calculateCanvasMapWidthAndHeight = width => {
    const x = width * 9;
    const y = x / 16;
    return {
      width: width,
      height: y
    }
}

const lineHover = lineHoverDataArr => {
  canvas.onmousemove = function(e) {
    lineHoverDataArr.forEach(twoNodesMetadata => {
      const { metadataTwoNode, transferMsg, line_id } = twoNodesMetadata;
      const { x1 ,y1, x2, y2 } = metadataTwoNode;
      const slope_rate = ((y2 - y1) / (x2 - x1)).toFixed(1);
      let rect = this.getBoundingClientRect(),
          x = e.clientX - rect.left,
          y = e.clientY - rect.top;
      let curr_slope_rate =  ((y - y1) / (x - x1)).toFixed(1);

      if (document.getElementById(`map_line_tooltip_${line_id}`)) {
        document.getElementById(`map_line_tooltip_${line_id}`).remove();
      }
      if (x > Math.min(x1, x2) && x < Math.max(x1, x2) && y > Math.min(y1, y2) && y < Math.max(y1, y2) && (Math.abs(curr_slope_rate) <= Math.abs(slope_rate) + 0.1 && Math.abs(curr_slope_rate) >= Math.abs(slope_rate) - 0.1)){
        const button = document.createElement("BUTTON");
        button.setAttribute("class", "bx--tooltip__trigger bx--tooltip--a11y bx--tooltip__trigger--definition bx--tooltip--bottom bx--tooltip--align-end");
        if (isDashboardPage()) {
          button.onclick = () => changeBandwidthLegendNodes('node_ips', metadataTwoNode);
        } else {
          button.style.cursor = 'default';
        }
        button.innerHTML = '&nbsp;';
        const tooltipTextDom = document.createElement("DIV");
        tooltipTextDom.setAttribute("class", "bx--assistive-text");
        tooltipTextDom.setAttribute("role", "tooltip");
        tooltipTextDom.innerHTML = transferMsg;
        const tooltipContainer = document.createElement("DIV");
        tooltipContainer.setAttribute("id", `map_line_tooltip_${line_id}`);
        tooltipContainer.setAttribute("class", "bx--tooltip--definition bx--tooltip--a11y");
        tooltipContainer.appendChild(button);
        tooltipContainer.appendChild(tooltipTextDom);
        tooltipContainer.style.position = "absolute";
        tooltipContainer.style.left = `${x}px`;
        tooltipContainer.style.top = `${y}px`;
        document.getElementById('new_dashboard_map').appendChild(tooltipContainer);
      }
    })
  };
}

const setMapBgImage = () => document.getElementById('new_dashboard_map').style.backgroundImage =  `url('${_map_img_path}')`;

const ifScreenResized = () => {
  jQuery(window).on('resize', () => {
    const mapMetadata = calculateCanvasMapWidthAndHeight(jQuery('#new_dashboard_map').width());
    updateActivityMapDisplay(mapMetadata.width, mapMetadata.height)
  });
}

const mainDriver = async () => {

  const currPage = getCurrPage();
  let updateChart;
  let charts;
  switch(currPage) {
    case pathMap.dashboard:
      setCanvasStyleWithMapDom();
      setMapBgImage();
      charts = await plotDashboardChart();
      ifScreenResized();
      updateChart = () => updateDashboardChart(null, charts);
      break;
    case pathMap.node_detail:
      charts = await plotNodeDetailChart();
      updateChart = () => updateNodeDetailChart(null, charts);
      break;
    case pathMap.all_bandwidth:
      charts = await plotNodeDetailChart(pathMap.all_bandwidth);
      updateChart = () => updateNodeDetailChart(pathMap.all_bandwidth, charts);
      break;
    case pathMap.session_detail:
      charts = await plotSessionDetailChart();
      updateChart = () => updateSessionDetailChart(charts);
      setTimeout(function(){ // set btns' status at first chart update
        update_btn_with_selected_transfer(job_step.id);
      }, dataUpdateFrequecySeconds * 1000);
      break;
    case pathMap.edit_map:
      setCanvasStyleWithMapDom();
      setMapBgImage();
      updateActivityMap();
      hideMapDisplayIfNotVisible(_fasp_node);
      ifScreenResized();
      updateChart = () => updateActivityMap();
      break;
    case pathMap.activity_map:
      setCanvasStyleWithMapDom();
      setMapBgImage();
      updateActivityMap();
      ifScreenResized();
      updateChart = () => updateActivityMap();
      break;
    default:
      charts = await plotDashboardChart();
      updateChart = () => updateDashboardChart(null, charts);
      console.log('no path match'); 
  }
  intervalFunc = updateChart && setInterval(function(){updateChart()}, dataUpdateFrequecySeconds * 1000);
}

mainDriver();
